from jinja2 import Undefined, DebugUndefined
from jinja2.utils import missing
from docxtpl import DocxTemplate
from jinja2 import Environment, BaseLoader
import jinja2.exceptions
from docassemble.base.util import DAFile

__all__ = ['CallAndDebugUndefined', 'get_jinja_errors']

class CallAndDebugUndefined(DebugUndefined):
    """Handles Jinja2 undefined errors by printing the name of the undefined variable.
    Extended to handle callable methods.
    """      
    def __call__(self, *pargs, **kwargs):
        return self
      
    def __getattr__(self, _: str) -> "CallAndDebugUndefined":
        return self

    __getitem__ = __getattr__  # type: ignore      
    

def get_jinja_errors(the_file:DAFile)->str:
  """Just try rendering the DOCX file as a Jinja2 template and catch any errors.
  Returns a string with the errors, if any.
  """
  env = Environment(loader=BaseLoader,undefined=CallAndDebugUndefined)
  
  doc = DocxTemplate(the_file.path())
  try: 
    doc.render({}, jinja_env=env)
  except jinja2.exceptions.TemplateSyntaxError as the_error:
    errmess = str(the_error)
    if hasattr(the_error, 'docx_context'):
      errmess += "\n\nContext:\n" + "\n".join(map(lambda x: "  " + x, the_error.docx_context))
    return errmess
  


import unittest
from .interview_generator import DAField, get_docx_variables, is_reserved_docx_label
from docx2python import docx2python
from pathlib import Path

class test_docxs(unittest.TestCase):

  def test_unmap_suffixes(self):
    unmap_suffixes_file = Path(__file__).parent / 'test/unmap_suffixes.docx'
    docx_data = docx2python(unmap_suffixes_file)
    text = docx_data.text
    all_vars = get_docx_variables(text)
    self.assertEqual(len(all_vars), 9, str(all_vars))
    self.assertIn('children[0].phone_number', all_vars)
    self.assertIn('children[1].birthdate', all_vars)
    self.assertIn('children[0].birthdate', all_vars)
    self.assertIn('children[0].birthdate', all_vars)
    self.assertIn('children[0].name.first', all_vars)
    self.assertIn('children[0].mailing_address.address', all_vars)
    self.assertIn('children[0].mailing_address.county', all_vars)
    self.assertIn('children[0].address.address', all_vars)
    self.assertIn('children[0].address.county', all_vars)
    self.assertIn('milkman.attorney.firm', all_vars)

  def test_reserved_docx_labels(self):
    reserved_labels_files = Path(__file__).parent / 'test/reserved_docx_variables.docx'
    docx_data = docx2python(reserved_labels_files)
    text = docx_data.text
    all_vars = get_docx_variables(text)
    reserved_labels = []
    for label in all_vars:
      if is_reserved_docx_label(label):
        reserved_labels.append(label)
    self.assertEqual(len(reserved_labels), 2, str(reserved_labels))
    self.assertIn('trial_court.address.county', reserved_labels)
    self.assertIn('users[0].address.address', reserved_labels)
    self.assertNotIn('users.address.zip', reserved_labels)


from typing import List, Union
from pathlib import Path
import yaml
from docassemble.base.util import log, DADict
from docassemble.base.core import objects_from_file
# TODO(brycew): is this too deep into DA? Unclear if there are options to write
# sources files to a package while it's running.
from docassemble.base.functions import package_data_filename

"""
Container for widely used data that may be altered by user inputs.
"""

__all__ = ['get_possible_deps_as_choices', 'get_pypi_deps_from_choices', 
           'get_yml_deps_from_choices']

# This is to workaround fact you can't do local import in Docassemble playground
class Object(object):
  pass

custom_values = Object()

# Var names of people-type lists that developers add while using the weaver.
# Could be a list, but a dict will match RESERVED_PLURALIZERS_MAP in the
# places it's used, so maybe easier to treat them both the same.
custom_values.people_plurals_map = {}

# Filled in by load_org_specific. Is a dictionary of configs that each org using
# the Weaver can set, per installation. The specific configs are:
# * jurisdiction_dependency_choices: a collection of other yaml files 
#   that the generated interview will include, specifically a jurisdiction
#   It's a map from keys, what the user will see when selecting dependencies, that maps to
#   a list of 3 items: the pypi dependency, the yaml include, and a boolean for default selection
custom_values.org_specific_config = None

def load_org_specific(all_custom_values=custom_values):
  if all_custom_values.org_specific_config is None:
    try:
      all_custom_values.org_specific_config = objects_from_file('org_specific.yml')
    except (FileNotFoundError, SystemError) as ex:
      # Populate some default values
      all_custom_values.org_specific_config = {
        'dependency_choices': {
          'Massachusetts State': 
              ['docassemble.ALMassachusetts>=0.0.7', 
               'docassemble.ALMassachusetts:al_massachusetts.yml',
               'jurisdiction',
               True],
          'MassAccess': 
              ['docassemble.MassAccess',  
               'docassemble.MassAccess:massaccess.yml',
               'organization',
               True]
        }
      }
      to_write = package_data_filename('data/sources/org_specific.yml')
      with open(to_write, 'w') as writ:
        writ.write(yaml.safe_dump(all_custom_values.org_specific_config))

def get_possible_deps_as_choices(dep_category=None, all_custom=custom_values):
  """Gets the possible yml files that the generated interview will depend on"""
  load_org_specific(all_custom)
  dep_choices = all_custom.org_specific_config['dependency_choices']
  if dep_category is None:
    return [{dep_key: dep_key, 'default': dep[3]} for dep_key, dep in dep_choices.items()]
  else:
    return [{dep_key: dep_key, 'default': dep[3]} for dep_key, dep in dep_choices.items() 
            if dep[2].lower() == dep_category.lower()]

def get_pypi_deps_from_choices(choices:Union[List[str], DADict],
    all_custom=custom_values):
  """Gets the Pypi dependency requirement (i.e. docassemble.AssemblyLine>=2.0.19)
  from some chosen dependencies"""
  return get_values_from_choices(choices, 0)

def get_yml_deps_from_choices(choices:Union[List[str], DADict],
    all_custom_values=custom_values):
  """Gets the yml file (i.e. docassemble.AssemblyLine:data/question/ql_baseline.yml)
  from some chosen dependencies"""
  return get_values_from_choices(choices, 1)

def get_values_from_choices(choices:Union[List[str], DADict], value_idx:int=0,
    all_custom_values=custom_values):
  load_org_specific(all_custom_values)
  if isinstance(choices, DADict):
    choice_list = choices.true_values()
  else: # List
    choice_list = choices
  
  return [all_custom_values.org_specific_config['dependency_choices'][chosen_val][value_idx] 
      for chosen_val in choice_list] 


from docassemble.webapp.playground import *
from docassemble.base.util import DAList


class DAQuestionList(DAList):
  def init(self, **kwargs):
    super().init(**kwargs)
    self.object_type = DAQuestion



import unittest
from .interview_generator import DAField

class test_fill_in_pdf_attributes(unittest.TestCase):

    def test_simple_pdf_field(self):
        pdf_field_tuple = ('field_name', 'default text', 0, [10, 10, 100, 30], '/Tx')
        new_field = DAField()
        new_field.fill_in_pdf_attributes(pdf_field_tuple)
        self.assertEqual(new_field.variable, 'field_name')
        self.assertEqual(new_field.final_display_var, 'field_name')
        self.assertEqual(new_field.has_label, True)
        self.assertEqual(new_field.field_type_guess, 'text')
        self.assertEqual(new_field.variable_name_guess, 'Field name')

    def test_date_field(self):
        pdf_field_tuple = ('birth_date', '', 0, [10, 10, 100, 30], '/Tx')
        new_field = DAField() 
        new_field.fill_in_pdf_attributes(pdf_field_tuple)
        self.assertEqual(new_field.variable, 'birth_date')
        self.assertEqual(new_field.final_display_var, 'birth_date')
        self.assertEqual(new_field.has_label, True)
        self.assertEqual(new_field.field_type_guess, 'date')
        self.assertEqual(new_field.variable_name_guess, 'Date of birth')

    def test_yes_text_field(self):
        pdf_field_tuple = ('has_ssn_yes', '', 0, [10, 10, 100, 30], '/Tx')
        new_field = DAField() 
        new_field.fill_in_pdf_attributes(pdf_field_tuple)
        self.assertEqual(new_field.variable, 'has_ssn_yes')
        self.assertEqual(new_field.final_display_var, 'has_ssn_yes')
        self.assertEqual(new_field.has_label, True)
        self.assertEqual(new_field.field_type_guess, 'yesno')
        self.assertEqual(new_field.variable_name_guess, 'Has ssn')

    def test_no_text_field(self):
        pdf_field_tuple = ('has_ssn_no', '', 0, [10, 10, 100, 30], '/Tx')
        new_field = DAField() 
        new_field.fill_in_pdf_attributes(pdf_field_tuple)
        self.assertEqual(new_field.variable, 'has_ssn_no')
        self.assertEqual(new_field.final_display_var, 'has_ssn_no')
        self.assertEqual(new_field.has_label, True)
        self.assertEqual(new_field.field_type_guess, 'yesno')
        self.assertEqual(new_field.variable_name_guess, 'Has ssn')

    def test_yesno_btn_field(self):
        pdf_field_tuple = ('has_ssn', '', 0, [10, 10, 100, 30], '/Btn')
        new_field = DAField() 
        new_field.fill_in_pdf_attributes(pdf_field_tuple)
        self.assertEqual(new_field.variable, 'has_ssn')
        self.assertEqual(new_field.final_display_var, 'has_ssn')
        self.assertEqual(new_field.has_label, True)
        self.assertEqual(new_field.field_type_guess, 'yesno')
        self.assertEqual(new_field.variable_name_guess, 'Has ssn')

    def test_sig_field(self):
        pdf_field_tuple = ('signature', '', 0, [10, 10, 100, 30], '/Sig')
        new_field = DAField() 
        new_field.fill_in_pdf_attributes(pdf_field_tuple)
        self.assertEqual(new_field.variable, 'signature')
        self.assertEqual(new_field.final_display_var, 'signature')
        self.assertEqual(new_field.has_label, True)
        self.assertEqual(new_field.field_type_guess, 'signature')
        self.assertEqual(new_field.variable_name_guess, 'Signature')


if __name__ == "__main__":
    unittest.main()


"""
Constants for interview_generator.py
"""

# This is to workaround fact you can't do local import in Docassemble playground
class Object(object):
  pass

generator_constants = Object()

# Words that are reserved exactly as they are
generator_constants.RESERVED_WHOLE_WORDS = [
  'signature_date',
  'docket_number',
  'user_needs_interpreter',
  'user_preferred_language',
]

# Prefixes for singular person-like objects, like trial courts that
# should be left undefined to trigger their question
generator_constants.UNDEFINED_PERSON_PREFIXES = [
  "trial_court",
]

# NOTE: if the suffix is allowed for a singular prefix, you must list the variable here
# E.g., we do not allow users.address.address in a DOCX, but we do allow trial_court.address.address
generator_constants.ALLOW_SINGULAR_SUFFIXES = [
  "trial_court"
]

# Prefixes as they would appear in a PDF (singular)
generator_constants.RESERVED_PREFIXES = [
  "users",
  "other_parties",
  "children",
  "plaintiffs",
  "defendants",
  "petitioners",
  "respondents",
  "spouses",
  "parents",
  "caregivers",
  "attorneys",
  "translators",
  "debt_collectors",
  "creditors",
  "witnesses",
  "courts",
  "signature_dates",
  "decedents",
  "interested_partys",
  "trial_court",
  "docket_numbers",
  "user",
  "other_party",
  "child",
  "plaintiff",
  "defendant",
  "petitioner",
  "respondent",
  "spouse",
  "parent",
  "caregiver",
  "attorney",
  "translator",
  "debt_collector",
  "creditor",
  "witness",
  "court",
  "signature_date",
  "decedent",
  "interested_party",
  "trial_court",
  "docket_numbers",
  # Can't find a way to make order not matter here
  # without making everything in general more messy
  "guardians_ad_litem",
  "guardian_ad_litem",
  "guardians",  
  "guardian",
  ]

generator_constants.RESERVED_PERSON_PLURALIZERS_MAP = {
  # Singular terms that get pluralized are deprecated
  "users": "users",
  "other_parties": "other_parties",
  "children": "children",
  "plaintiffs": "plaintiffs",
  "defendants": "defendants",
  "petitioners": "petitioners",
  "respondents": "respondents",
  "spouses": "spouses",
  "parents": "parents",
  "caregivers": "caregivers",
  "attorneys": "attorneys",
  "translators": "translators",
  "debt_collectors": "debt_collectors",
  "creditors": "creditors",
  "witnesses": "witnesses",
  "courts": "courts",
  "signature_date": "signature_date",
  "decedents": "decedents",
  "interested_parties": "interested_parties",
  "guardians_ad_litem": "guardians_ad_litem",
  "guardians": "guardians",
  # These are left in for backwards compatibility
  'user': 'users',
  'plaintiff': 'plaintiffs',
  'defendant': 'defendants',
  'petitioner': 'petitioners',
  'respondent': 'respondents',
  'spouse': 'spouses',
  'parent': 'parents',
  'guardian': 'guardians',
  'caregiver': 'caregivers',
  'attorney': 'attorneys',
  'translator': 'translators',
  'debt_collector': 'debt_collectors',
  'creditor': 'creditors',
  # Non "s" plurals
  'other_party': 'other_parties',
  'child': 'children',
  'guardian_ad_litem': 'guardians_ad_litem',
  'witness': 'witnesses',
  'decedent': 'decedents',
  'interested_party': 'interested_parties',
}

generator_constants.RESERVED_PRIMITIVE_PLURALIZERS_MAP = {
  'docket_numbers': 'docket_numbers' 
}

generator_constants.RESERVED_PLURALIZERS_MAP = {
  **generator_constants.RESERVED_PERSON_PLURALIZERS_MAP, 
  **generator_constants.RESERVED_PRIMITIVE_PLURALIZERS_MAP,
  **{
    'court': 'courts', # for backwards compatibility 
  }
}

# Any reason to not make all suffixes available to everyone?
# Yes: it can break variables that overlap but have a different meaning
# Better to be explicit

# Some common attributes that can be a clue it's a person object
generator_constants.PEOPLE_SUFFIXES_MAP = {
  '_name': "",  # full name
  '_name_full': "",  # full name
  '_name_first': ".name.first",
  '_name_middle': ".name.middle",
  '_name_middle_initial': ".name.middle_initial()",
  '_name_last': ".name.last",
  '_name_suffix': ".name.suffix",
  '_gender': ".gender",
  '_gender_male': ".gender_male",
  '_gender_female': ".gender_female",
  '_gender_nonbinary': ".gender_nonbinary",
  '_gender_undisclosed': ".gender_undisclosed",
  '_gender_self_described': ".gender_self_described",
  '_birthdate': ".birthdate.format()",
  '_age': ".age_in_years()",
  '_email': ".email",
  '_phone': ".phone_number",
  '_phone_number': ".phone_number",
  '_mobile': ".mobile_number",
  '_mobile_number': ".mobile_number",
  '_phones': ".phone_numbers()",
  '_address_block': ".address.block()",
  # TODO: deprecate street and street2 from existing forms and documentation
  '_address_street': ".address.address",
  '_address_street2': ".address.unit",
  '_address_address': ".address.address",
  '_address_unit': ".address.unit",
  '_address_city': ".address.city",
  '_address_state': ".address.state",
  '_address_zip': ".address.zip",
  '_address_county': ".address.county",
  '_address_country': ".address.country",
  '_address_on_one_line': ".address.on_one_line()",
  '_address_line_one': ".address.line_one()",
  '_address_city_state_zip': ".address.line_two()",
  '_signature': ".signature",
  '_mailing_address_block': ".mailing_address.block()",
  '_mailing_address_street': ".mailing_address.address",
  '_mailing_address_street2': ".mailing_address.unit",
  '_mailing_address_address': ".mailing_address.address",
  '_mailing_address_unit': ".mailing_address.unit",
  '_mailing_address_city': ".mailing_address.city",
  '_mailing_address_state': ".mailing_address.state",
  '_mailing_address_zip': ".mailing_address.zip",
  '_mailing_address_county': ".mailing_address.county",
  '_mailing_address_country': ".mailing_address.country",
  '_mailing_address_on_one_line': ".mailing_address.on_one_line()",
  '_mailing_address_line_one': ".mailing_address.line_one()",
  '_mailing_address_city_state_zip': ".mailing_address.line_two()",
  '_mailing_address': ".mailing_address",
}

generator_constants.PEOPLE_SUFFIXES = list(generator_constants.PEOPLE_SUFFIXES_MAP.values()) + ['.name.full()','.name']

# reserved_suffixes_map
generator_constants.RESERVED_SUFFIXES_MAP = {**generator_constants.PEOPLE_SUFFIXES_MAP, **{
  # Court-specific
  # '_name_short': not implemented,
  '_division': ".division",
  '_county': ".address.county",
  '_department': ".department",
}}

# these might be used in a docx, but we don't transform PDF fields to use these
# suffixes
generator_constants.DOCX_ONLY_SUFFIXES = [
    r'\.birthdate',
    r'\.birthdate.format\(.*\)',
    r'\.familiar\(\)',
    r'\.familiar_or\(\)',
    r'\.phone_numbers\(\)',
    r'\.formatted_age\(\)'
]

generator_constants.DISPLAY_SUFFIX_TO_SETTABLE_SUFFIX = {
  '\.address.block\(\)$': '.address.address',
  '\.address.line_one\(\)$': '.address.address',
  '\.address.line_two\(\)$': '.address.address',
  '\.address.on_one_line\(\)$': '.address.address',
  '\.age_in_years\(\)$': '.birthdate',
  '\.birthdate.format\(.*\)$': '.birthdate',
  '\.familiar_or\(\)$': '.name.first',
  '\.familiar\(\)$': '.name.first',
  '\.formatted_age\(.*\)$': '.birthdate',
  '\.mailing_address.block\(\)$': '.mailing_address.address',
  '\.mailing_address.line_one\(\)$': '.mailing_address.address',
  '\.mailing_address.line_two\(\)$': '.mailing_address.address',
  '\.mailing_address.on_one_line\(\)$': '.mailing_address.address',
  '\.name.middle_initial\(\)$': '.name.first',
  '\.phone_numbers\(\)$': '.phone_number',
}

# Test needed: Jinja `{{ parents[0].name_of_dog }}` should remain the same,
# not `.name.full()` in the review screen displayed value
generator_constants.FULL_DISPLAY = {
  '\.name$': '.name.full()',
  '\.address$': '.address.block()',
  '\.mailing_address$': '.mailing_address.block()'
}

# Possible values for 'Allowed Courts', when looking up courts to submit to
generator_constants.COURT_CHOICES = [
  'Boston Municipal Court',
  'District Court',
  'Superior Court',
  'Housing Court',
  'Probate and Family Court',
  'Juvenile Court',
  'Land Court'
]